#!/usr/bin/perl -w

# Copyright 2014, 2016 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use List::Util 'max','min';

use FindBin;
use lib::abs "$FindBin::Bin/../tools";
use MyDragon ':all';
use MyLineSegments;

# uncomment this to run the ### lines
# use Smart::Comments;

{
  # counting distinct shapes

  my $four_sides = 1;

  # length 1 = 2
  #          *
  #          |
  # *----*   *
  #
  # length 2 = 6
  #
  # *
  # |
  # *  *---*  *---*                 *  *
  # |  |          |                 |  |
  # *  *          *  *---*---*  *---*  *---*
  #
  # count shapes which
  #    - never traverse a segment more than once
  # 2,6,22,88,372,1628,7312         A096267
  #    - traverse all sides of an enclosed unit square
  # 1 2 3  4  5   6    7
  # 2,6,22,88,372,1626              2*A121220 hexagonal
  # 2,6,22,88,372,1626,7292         not
  #
  my $dummy = <<HERE;
(search-forward
"*---*---*
|       |
*---*---*")
HERE
  require Clone;

  my %seen;
  my @elems;
  push @elems, [MyLineSegments->new];

  my @values;

  foreach my $len (1 .. 8) {
    ### @elems
    my $skip_sides = 0;
    my $total_boundary = 0;
    my $total_area = 0;
    my @new_elems;
    foreach my $elem (@elems) {
      my ($segs) = @$elem;
      my @points = $segs->points;
      if (! @points) { @points = ([0,0]); }

      foreach my $point (@points) {
        my ($x,$y) = @$point;

        foreach my $dir4 (0 .. $#dir4_to_dx) {
          my $dx = $dir4_to_dx[$dir4];
          my $dy = $dir4_to_dy[$dir4];
          ### try dir: "$dx,$dy"

          my $x2 = $x + $dx;
          my $y2 = $y + $dy;
          ### $segs
          if ($segs->is_traversed($x,$y, $x2,$y2)) {
            ### already traversed ...
            next;
          }
          ### not already traversed ...

          if ($four_sides) {
            # only 4-sides of enclosed
            if ($segs->is_visited($x2,$y2)) {
              my ($lx,$ly) = xy_rotate_plus90($dx,$dy);
              my $c = ($segs->is_traversed($x,$y, $x+$lx,$y+$ly) ? 1 : 0)
                + ($segs->is_traversed($x2,$y2, $x2+$lx,$y2+$ly) ? 1 : 0)
                  + ($segs->is_traversed($x+$lx,$y+$ly, $x2+$lx,$y2+$ly) ? 1 : 0);
              unless ($c == 0 || $c == 3) {
                $skip_sides++;
                next;
              }
              my ($rx,$ry) = xy_rotate_minus90($dx,$dy);
              $c = ($segs->is_traversed($x,$y, $x+$rx,$y+$ry) ? 1 : 0)
                + ($segs->is_traversed($x2,$y2, $x2+$rx,$y2+$ry) ? 1 : 0)
                  + ($segs->is_traversed($x+$rx,$y+$ry, $x2+$rx,$y2+$ry) ? 1 : 0);
              unless ($c == 0 || $c == 3) {
                $skip_sides++;
                next;
              }
            }
          }

          my $new_segs = Clone::clone($segs);
          $new_segs->add($x,$y, $x2,$y2);
          my ($xmin,$ymin, $xmax,$ymax) = $new_segs->bbox;
          if ($xmin < 0 || $ymin < 0) {
            $new_segs->translate(max(0,-$xmin), max(0,-$ymin));
          }
          my $key = $new_segs->segments_string;
          ### $key
          next if $seen{$key}++;

          push @new_elems, [$new_segs];

          my $boundary = segs_boundary_length($new_segs);
          $total_boundary += $boundary;
          my $area = segs_area($new_segs);
          $total_area += $area;

          if ($four_sides) {
            if (4*$area + $boundary != 2*$len) {
              die "oops, not 4A+B=2N";
            }
          }

          if ($len == 1) {
            print "  new $key  boundary=$boundary area=$area\n";
            $new_segs->print_ascii;
          }
        }
      }
    }
    @elems = @new_elems;
    my $value = scalar(@elems);
    print "len=$len  $value   skip sides=$skip_sides total boundary=$total_boundary total area=$total_area\n";

    push @values, $total_area;
    # push @values, $total_boundary/4;
    # push @values, $value;
  }
  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(array => \@values, verbose=>1);
  exit 0;

  sub segs_boundary_length {
    my ($segs) = @_;
    ### segs_boundary_length() ...

    my ($x_origin,$y_origin) = segs_origin($segs);
    my ($dx_origin, $dy_origin);
    if ($segs->is_traversed($x_origin,$y_origin, $x_origin+1,$y_origin)) {
      $x_origin++;
      $dx_origin = 1;
      $dy_origin = 0;
    } elsif ($segs->is_traversed($x_origin,$y_origin, $x_origin,$y_origin+1)) {
      $y_origin++;
      $dx_origin = 0;
      $dy_origin = 1;
    } else {
      die "Oops, initial segment not right or up";
    }
    my ($x,$y) = ($x_origin,$y_origin);
    my ($dx,$dy) = ($dx_origin,$dy_origin);

    ### initial: "$x,$y  $dx,$dy"

    my $length = 0;
  OUTER: for (;;) {
      $length++;
      ($dx,$dy) = xy_rotate_minus90($dx,$dy);
      foreach (1 .. 4) {
        my $x2 = $x + $dx;
        my $y2 = $y + $dy;
        ### try: "direction $dx,$dy"
        if ($segs->is_traversed($x,$y, $x2,$y2)) {
          $x = $x2;
          $y = $y2;
          if ($x == $x_origin && $y == $y_origin
              && $dx == $dx_origin && $dy == $dy_origin) {
            return $length;
          }
          next OUTER;
        }
        ($dx,$dy) = xy_rotate_plus90($dx,$dy);
      }
      die "oops, no next direction";
    }
  }

  sub segs_origin {
    my ($segs) = @_;
    my ($x_min,$y_min, $x_max,$y_max) = $segs->bbox;
    foreach my $x ($x_min .. $x_max) {
      if ($segs->is_visited($x,$y_min)) {
        return ($x,$y_min);
      }
    }
    die "Oops, no point of y_min visited";
  }

  sub segs_area {
    my ($segs) = @_;
    my ($x_min,$y_min, $x_max,$y_max) = $segs->bbox;

    my %outside;
    foreach my $y ($y_min .. $y_max-1) {
      $outside{$x_min-1}{$y} = 1;
      $outside{$x_max}{$y} = 1;
    }
    foreach my $x ($x_min .. $x_max-1) {
      $outside{$x}{$y_min-1} = 1;
      $outside{$x}{$y_max} = 1;
    }

    my $more;
    do {
      $more = 0;
      foreach my $y ($y_min .. $y_max-1) {
        foreach my $x ($x_min .. $x_max-1) {  # rightwards
          if ($outside{$x-1}{$y}
              && ! $outside{$x}{$y}
              && ! $segs->is_traversed($x,$y, $x,$y+1)) {
            $outside{$x}{$y} = 1;
            $more = 1;
          }
        }
        foreach my $x (reverse $x_min+1 .. $x_max) { # leftwards
          if (! $outside{$x-1}{$y}
              && $outside{$x}{$y}
              && ! $segs->is_traversed($x,$y, $x,$y+1)) {
            $outside{$x-1}{$y} = 1;
            $more = 1;
          }
        }
      }

      foreach my $x ($x_min .. $x_max-1) {
        foreach my $y ($y_min .. $y_max-1) {  # upwards
          if ($outside{$x}{$y-1}
              && ! $outside{$x}{$y}
              && ! $segs->is_traversed($x,$y, $x+1,$y)) {
            $outside{$x}{$y} = 1;
            $more = 1;
          }
        }
        foreach my $y (reverse $y_min+1 .. $y_max) { # downwards
          if ($outside{$x}{$y}
              && ! $outside{$x}{$y-1}
              && ! $segs->is_traversed($x,$y, $x+1,$y)) {
            $outside{$x}{$y-1} = 1;
            $more = 1;
          }
        }
      }
    } while ($more);

    my $area = 0;
    foreach my $y ($y_min .. $y_max-1) {
      foreach my $x ($x_min .. $x_max-1) {
        if (! $outside{$x}{$y}) {
          $area++;
        }
      }
    }
    return $area;
  }
}



{
  # counting distinct shapes, fixed origin so translates differ

  # length 1 = 1
  # *----*
  #
  # length 2 = 3
  #                                   *
  #                                   |
  # *----*     *----*----*       *----*
  #      |
  #      *
  #
  # length 3 = 9
  #                                    *----*
  #                                    |
  # O----*     O----*----*        O----*
  #      |               |
  # *----*               *             *
  #                                    |
  # O----*     O----*----*----*        *
  #      |                             |
  #      *               *        O----*            
  #      |               |
  #      *     O----*----*        *----*
  #                                    |
  # O----*                        O----*
  #      |
  #      *----*
  #
  # length 4
  #   +2    +3    +3
  #   +3    +3    +3
  #   +3    +3    +2     = 25
  #
  # count paths which
  #    - start East
  #    - never re-visit their start point
  #    - never traverse a segment more than once
  #  1,3,9,25,73,203,576,1589,4446,12218
  #
  #---------------------------------------------------
  # count paths which
  #    - start East
  #    - never traverse a segment more than once
  #  1,3,9,27,77,221,621,1750,4863,13542
  #
  #---------------------------------------------------
  # count paths which
  #    - start East
  #    - never traverse a segment more than once
  #    - traverse all sides of an enclosed unit square, for all sub-lengths
  #  1 2 3  4  5   6
  #  1,3,9,27,77,219,617,1732,4814,13374
  #    - also never revisit the start
  #  1,3,9,25,73,203,576,1589,4443,12211
  #
  require Clone;
  my %seen;
  my @elems;
  {
    my $segs = MyLineSegments->new;
    $segs->add(0,0, 1,0);
    push @elems, [ 1,0, $segs ];
  }

  my @values;
  push @values, scalar(@elems);

  foreach my $len (2 .. 10) {
    ### @elems
    my @new_elems;
    foreach my $elem (@elems) {
      my ($x, $y, $segs) = @$elem;
      foreach my $dir4 (0 .. $#dir4_to_dx) {
        my $dx = $dir4_to_dx[$dir4];
        my $dy = $dir4_to_dy[$dir4];
        ### try dir: "$dx,$dy"

        my $x2 = $x + $dx;
        my $y2 = $y + $dy;
        next if $segs->is_traversed($x,$y, $x2,$y2);

         next if $x2 == 0 && $y2 == 0;

        if ($segs->is_visited($x2,$y2)) {
          my ($lx,$ly) = xy_rotate_plus90($dx,$dy);
          if ($segs->is_traversed($x,$y, $x+$lx,$y+$ly)
              && $segs->is_traversed($x2,$y2, $x2+$lx,$y2+$ly)
              && ! $segs->is_traversed($x+$lx,$y+$ly, $x2+$lx,$y2+$ly)) {
            next;
          }            
          my ($rx,$ry) = xy_rotate_minus90($dx,$dy);
          if ($segs->is_traversed($x,$y, $x+$rx,$y+$ry)
              && $segs->is_traversed($x2,$y2, $x2+$rx,$y2+$ry)
              && ! $segs->is_traversed($x+$rx,$y+$ry, $x2+$rx,$y2+$ry)) {
            next;
          }            
        }

        my $new_segs = Clone::clone($segs);
        ### $new_segs
        $new_segs->add($x,$y, $x2,$y2);
        my $key = $new_segs->segments_string;
        ### $key
        next if $seen{$key}++;

        push @new_elems, [$x2,$y2, $new_segs];
      }
    }
    @elems = @new_elems;
    my $value = scalar(@elems);
    push @values, $value;
    print "len=$len  $value\n";
  }
  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(array => \@values, verbose=>1);
  exit 0;
}



{
  # counting distinct path coordinates, so segment order matters

  #
  # count paths which
  #    - start East
  #    - never traverse a segment more than once
  # 1,3,9,27,79,229,657,1875,5317,15023
  #    - never re-visit their start point
  # 1,3,9,25,75,211,609,1703,4853,13531
  #
  #---------------------------------------------------
  # count paths which
  #    - start East
  #    - never traverse a segment more than once
  #    - traverse all sides of an enclosed unit square, for all sub-lengths
  #  1 2 3  4  5   6
  #  1,3,9,27,79,227,649,1841,5185,14569
  #    - never re-visit their start point
  #  1,3,9,25,75,211,605,1687,4777,13279
  #
  #---------------------------------------------------
  # count paths which
  #    - start East
  #    - never traverse a segment more than once
  #    - don't cross themselves
  #  1,3,9,27,79,221,609,1670,4557,12357
  #
  require Clone;
  my %seen;
  my @elems;
  {
    my $segs = MyLineSegments->new;
    $segs->add(0,0, 1,0);
    push @elems, [ 1,0, $segs, "0,0 to 1,0" ];
  }

  my @values;
  push @values, scalar(@elems);

  foreach my $len (2 .. 10) {
    ### @elems
    my @new_elems;
    foreach my $elem (@elems) {
      my ($x, $y, $segs, $str, $stop) = @$elem;
      next if $stop;
      foreach my $dir4 (0 .. $#dir4_to_dx) {
        my $dx = $dir4_to_dx[$dir4];
        my $dy = $dir4_to_dy[$dir4];
        ### try dir: "$dx,$dy"

        my $x2 = $x + $dx;
        my $y2 = $y + $dy;
        next if $segs->is_traversed($x,$y, $x2,$y2);

        # next if $x2 == 0 && $y2 == 0;

        # if ($segs->is_visited($x2,$y2)) {
        #   my ($lx,$ly) = xy_rotate_plus90($dx,$dy);
        #   if ($segs->is_traversed($x,$y, $x+$lx,$y+$ly)
        #       && $segs->is_traversed($x2,$y2, $x2+$lx,$y2+$ly)
        #       && ! $segs->is_traversed($x+$lx,$y+$ly, $x2+$lx,$y2+$ly)) {
        #     next;
        #   }            
        #   my ($rx,$ry) = xy_rotate_minus90($dx,$dy);
        #   if ($segs->is_traversed($x,$y, $x+$rx,$y+$ry)
        #       && $segs->is_traversed($x2,$y2, $x2+$rx,$y2+$ry)
        #       && ! $segs->is_traversed($x+$rx,$y+$ry, $x2+$rx,$y2+$ry)) {
        #     next;
        #   }            
        # }

        # non-crossing ...
        if ($segs->is_visited($x2,$y2)) {
          my ($lx,$ly) = xy_rotate_plus90($dx,$dy);
          if ($segs->is_traversed($x2,$y2, $x2+$lx,$y2+$ly)
              && $segs->is_traversed($x2,$y2, $x2-$lx,$y2-$ly)) {
            $stop = 1;
          }
        }

        my $new_str = "$str to $x2,$y2";
        my $new_segs = Clone::clone($segs);
        $new_segs->add($x,$y, $x2,$y2);

        push @new_elems, [$x2,$y2, $new_segs, $new_str, $stop];
      }
    }
    @elems = @new_elems;
    my $value = scalar(@elems);
    push @values, $value;
    print "len=$len  $value\n";
  }
  require Math::OEIS::Grep;
  Math::OEIS::Grep->search(array => \@values, verbose=>1);
  exit 0;
}
