#!/usr/bin/perl -w

# Copyright 2016, 2017, 2018, 2019, 2020, 2021 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.004;
use strict;
use Math::PlanePath::ComplexMinus 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyLineSegments;
use MyPlanar ':all';
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 8;
if (@ARGV) { $level = shift @ARGV; }

print "  % generated by tools/$FindBin::Script $level\n";
print "  \\newcommand{\\MyLevel}{$level}\n";

my $path = Math::PlanePath::ComplexMinus->new;
my ($n_lo, $n_hi) = $path->level_to_n_range($level);


sub points_square {
  my ($x,$y) = @_;
  return [ [2*$x-1, 2*$y-1],
           [2*$x+1, 2*$y-1],
           [2*$x+1, 2*$y+1],
           [2*$x-1, 2*$y+1] ];
}

# all segments of the plain i-1 shape
my $plain_segments  = MyLineSegments->new (name => 'plain');

# all segments of the i-1 shape rotated 180 degrees, and which will be
# reduced to just those not in the plain segments
my $other_segments = MyLineSegments->new (name => 'other');

# segments around negatable points of the plain i-1 shape, meaning their 180
# rotation is still in the shape
my $negatable_segments = MyLineSegments->new (name => 'negatable');

# unit squares of the negatable points of the plain i-1 shape
my $negatable_gpc = MyPlanar::points_to_gpc(points_square(0,0));

my $MinusNegA = 0;
my $MinusNegNonA = 0;
foreach my $n ($n_lo .. $n_hi) {
  my ($x,$y) = $path->n_to_xy($n);
  my $points = points_square($x,$y);
  $plain_segments->add_cycle(map {@$_} @$points);

  if ($path->xy_to_n(-$x,-$y) <= $n_hi) {
    $negatable_segments->add_cycle(map {@$_} @$points);
    my $gpc = MyPlanar::points_to_gpc($points);
    $negatable_gpc = Math::Geometry::Planar::GpcClip('UNION',$negatable_gpc,$gpc);
    $MinusNegA++;
  } else {
    $MinusNegNonA++;
  }

  ($x,$y) = (-$x,-$y);
  $points = points_square($x,$y);
  $other_segments->add_cycle(map {@$_} @$points);
}

# Remove $negatable_segments, they will be a filled area.
# By symmetry, $other_segments also contains $negatable_segments, which is
# not true for instance in MinusRot.
#
$negatable_segments->is_subset_of($plain_segments) or die;
$negatable_segments->is_subset_of($other_segments) or die;
$plain_segments->remove_segments($negatable_segments);
$other_segments->remove_segments($negatable_segments);

# remove $plain_segments, so $other_segments is rest of negated copy
$other_segments->remove_segments($plain_segments);
$plain_segments->is_disjoint($other_segments) or die;

my $MinusNegB = 0;
{
  my @points_list;
  my @contours = Math::Geometry::Planar::Gpc2Polygons($negatable_gpc);
  my $total_holes = 0;
  foreach my $contour (@contours) {  # Math::Geometry::Planar contour
    ### $contour
    # Second and subsequent polygon of the contour are holes.
    # Think their points already cycle the opposite way as required by pgf
    # default \fill[nonzero rule], but using \fill[even odd rule] works either
    # way.
    $Printwrap = 2;
    printwrap(" \\fill[my negatable,even odd rule]");
    $Printwrap += 2;
    my $polygons = $contour->polygons;  # arrayref of points arrayrefs
    $total_holes += scalar(@$polygons)-1;
    my $separator = '';
    foreach my $points (@$polygons) {
      ### $points
      printwrap($separator);
      $separator = "\n";
      printwrap_poly($points);
      push @points_list, @$points;
    }
    printwrap(";\n");
    $MinusNegB += $contour->perimeter / 2;
  }
  $total_holes == 0 or die;
}

### $plain_segments
$Printwrap = 2;
printwrap("\\draw[my plain line] ");
$Printwrap += 2;
$plain_segments->printwrap;
printwrap(";\n");

$Printwrap = 2;
printwrap("\\draw[my other line] ");
$Printwrap += 2;
$other_segments->printwrap;
printwrap(";\n");

print "% GP-Test  MinusNegA($level) == $MinusNegA\n";
print "% GP-Test  MinusNegNonA($level) == $MinusNegNonA\n";
print "% GP-Test-Last  MinusNegB_by_pred($level) == $MinusNegB\n";
print "  \\newcommand{\\MyMinusNegA}{$MinusNegA}\n";
print "  \\newcommand{\\MyMinusNegNonA}{$MinusNegNonA}\n";
print "  \\newcommand{\\MyMinusNegB}{$MinusNegB}\n";
print "  % end generated\n";
exit 0;









__END__


# use 5.004;
# use strict;
# use Math::PlanePath::ComplexMinus 117; # v.117 for level_to_n_range()
#
# use FindBin;
# use lib $FindBin::Bin;
# use MyLineSegments;
# use MyPrintwrap ':all';
#
# # uncomment this to run the ### lines
# # use Smart::Comments;
#
#
# my $level = 8;
# if (@ARGV) {
#   $level = shift @ARGV;
# }
# print "  % generated by tools/$FindBin::Script\n";
# print "  \\newcommand{\\MyLevel}{$level}\n";
#
# my $path = Math::PlanePath::ComplexMinus->new;
# my ($n_lo, $n_hi) = $path->level_to_n_range($level);
#
#
# my $n_segments = MyLineSegments->new (name => 'n');
# my $other_segments = MyLineSegments->new (name => 'neg');
# my $negatable_segments = MyLineSegments->new (name => 'negatable');
#
# sub xy_to_type {
#   my ($x,$y) = @_;
#   my $n   = $path->xy_to_n($x,$y);
#   my $neg = $path->xy_to_n(-$x,-$y);
#   if ($n <= $n_hi) {
#     return ($neg <= $n_hi ? 'negatable' : 'n');
#   } else {
#     return ($neg <= $n_hi ? 'neg' : '');
#   }
# }
#
# sub boundary_to_segs {
#   my ($a,$b) = @_;
#   if ($a eq 'negatable') {
#     if ($b eq 'negatable') { return; }                       # hollow
#     else              { return $negatable_segments; }   # boundary
#   }
#   if ($a eq 'n') {
#     if ($b eq 'negatable') { return; }
#     return $n_segments;
#   }
#   if ($a eq 'neg') {
#     if ($b eq 'n')    { return; }                       # n--neg as "n"
#     if ($b eq 'negatable') { return; }
#     return $other_segments;
#   }
#   die;
# }
#
# my %negatable;
#
# foreach my $n ($n_lo .. $n_hi) {
#   my ($x,$y) = $path->n_to_xy($n);
#   foreach my $neg (0,1) {
#     my $a = xy_to_type($x,$y);
#     if ($a eq 'negatable') {
#       $negatable{"$x,$y"} = 1;
#     }
#     ### at: "$x,$y  a=$a (neg=$neg)"
#     foreach my $segments (boundary_to_segs($a, xy_to_type($x,$y-1))) {
#       $segments->add(2*$x-1,2*$y-1, 2*$x+1,2*$y-1);   # S
#     }
#     foreach my $segments (boundary_to_segs($a, xy_to_type($x,$y+1))) {
#       $segments->add(2*$x-1,2*$y+1, 2*$x+1,2*$y+1);   # N
#     }
#     foreach my $segments (boundary_to_segs($a, xy_to_type($x-1,$y))) {
#       $segments->add(2*$x-1,2*$y-1, 2*$x-1,2*$y+1);   # W
#     }
#     foreach my $segments (boundary_to_segs($a, xy_to_type($x+1,$y))) {
#       ### E segment: $segments->{'name'}
#       $segments->add(2*$x+1,2*$y-1, 2*$x+1,2*$y+1);   # E
#     }
#     $x = -$x;
#     $y = -$y;
#   }
# }
#
# $Printwrap = 2;
# printwrap("\\fill[my intersect] ");
# $Printwrap += 2;
# $negatable_segments->printwrap;
# printwrap(";\n");
#
# $Printwrap = 2;
# printwrap("\\draw[my squares line] ");
# $Printwrap += 2;
# $n_segments->printwrap;
# printwrap(";\n");
#
# $Printwrap = 2;
# printwrap("\\draw[my other line] ");
# $Printwrap += 2;
# $other_segments->printwrap;
# printwrap(";\n");
#
# my $MinusNegA = scalar keys %negatable;
# print "% GP-Test  MinusNegA($level) == $MinusNegA\n";
# print "  \\newcommand{\\MyMinusNegA}{$MinusNegA}\n";
# print "  % end generated\n";
# exit 0;
