#!/usr/bin/perl -w

# Copyright 2014, 2015, 2016, 2017, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.004;
use strict;
use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()
use Math::PlanePath::ComplexMinus 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyDragon 'PlusOffsetJ_xy','PlusToDragon',
  'xy_rotate_plus90','xy_rotate_minus90','div','rect_corners_sort';
use MyLineSegments;
use MyPrintwrap ':all';

my $level = 8;  # complex base level = twindragon level
my $direction = 'straight';
if (@ARGV) { $direction = $ARGV[0]; }

print "  % generated by tools/$FindBin::Script $direction\n";
print "  \\newcommand{\\MyLevel}{$level}\n";
print "  \\newcommand{\\MyLevelNext}{",$level+1,"}\n";

{
  my $path = Math::PlanePath::ComplexMinus->new;
  my $transform = sub {
    my ($x,$y) = @_;
    return (2*$x, 2*$y);
  };

  my ($n_lo, $n_hi) = $path->level_to_n_range($level);
  my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);

  my ($jx,$jy)   = PlusOffsetJ_xy($level);
  my ($j2x,$j2y) = PlusOffsetJ_xy($level+1);
  if ($level % 2 == 0) {
    $jx = -$jx;
    $j2x = -$j2x;
  } else {
    $jy = -$jy;
    $j2y = -$j2y;
  }
  my ($ox,$oy) = ($j2x-$jx, $j2y-$jy);
  my $os = $ox + $oy;
  my $od = $ox - $oy;

  my %squares;
  my $segments = MyLineSegments->new;
  foreach my $n ($n_lo .. $n_hi) {
    my ($x,$y) = $path->n_to_xy($n);
    my ($tx,$ty) = $transform->($x,$y);

    my $want_square;
    if ($direction eq 'straight') {
      $want_square = ($x == 0 || $y == 0 || $x == $ox || $y == $oy);
    } else { # $direction eq 'diagonal'
      my $s = $x+$y;
      my $d = $x-$y;
      $want_square = ($s == 0 || $d == 0 || $s == $os || $d == $od);
    }
    if ($want_square) {
      $squares{"$tx,$ty"} = 1;
    }

    $segments->add($tx-1,$ty-1, $tx+1,$ty-1); # horizontal
    $segments->add($tx-1,$ty+1, $tx+1,$ty+1);
    $segments->add($tx-1,$ty-1, $tx-1,$ty+1); # vertical
    $segments->add($tx+1,$ty-1, $tx+1,$ty+1);
  }

  {
    $Printwrap = 2;
    printwrap("\\fill[my square colour] foreach \\p in {");
    $Printwrap += 2;
    foreach my $txty (sort keys %squares) {
      printwrap("($txty),");
    }
    printwrap("}\n");
    printwrap("{ \\p +(-1+\\MyMargin,-1+\\MyMargin) rectangle +(1-\\MyMargin,1-\\MyMargin) };\n");
  }

  $Printwrap = 2;
  printwrap("\\draw ");
  $Printwrap += 2;
  $segments->printwrap;
  printwrap(";\n");

  {
    my ($x,$y) = $transform->(0,0);
    print "  \\node (minus-start) at ($x,$y) [my dot];\n";
  }
  {
    my ($x,$y) = $transform->($x_hi,$y_hi);
    print "  \\node (minus-end) at ($x,$y) [my dot];  % untransformed $ox,$oy\n";
    # ($ox == $x_hi && $oy == $y_hi)
    #   or die "oxy=$ox,$oy hi=$x_hi,$y_hi";
    print "  \\newcommand\\MyMinusNhi{$n_hi}\n";
    print "% GP-Test  MinusPoint($n_hi) == $x_hi+I*$y_hi\n";
  }
}

{
  print "  \\begin{scope}[shift={(pos-2)}]\n";
  my $path = Math::PlanePath::DragonCurve->new;
  my $transform = sub {
    my ($x,$y) = @_;
    # rotate -135 and scale up
    ($x,$y) = div($x,$y);
    ($x,$y) = xy_rotate_minus90($x,$y);
    return ($x,$y);
  };

  my ($n_lo, $n_hi) = $path->level_to_n_range($level+1);
  my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);

  {
    my ($x,$y) = $transform->(0,0);
    print "  \\node (twindragon-start) at ($x,$y) [my dot];\n";
  }
  {
    my ($x,$y) = $transform->($x_hi,$y_hi);
    print "  \\node (twindragon-end) at ($x,$y) [my dot];\n";
  }

  my ($jx,$jy) = PlusOffsetJ_xy($level);
  my ($jdragon_x,$jdragon_y) = PlusToDragon($jx,$jy);
  print "% GP-Test  PlusOffsetJ($level) == $jx+I*$jy\n";
  print "% GP-Test  PlusToDragon($jx+I*$jy) == $jdragon_x+I*$jdragon_y\n";
  my $jdragon_s = $jdragon_x + $jdragon_y;
  my $jdragon_d = $jdragon_x - $jdragon_y;

  my ($j2x,$j2y) = PlusOffsetJ_xy($level+1);
  my ($j2dragon_x,$j2dragon_y) = PlusToDragon($j2x,$j2y);
  print "% GP-Test  PlusOffsetJ($level+1) == $j2x+I*$j2y\n";
  print "% GP-Test  PlusToDragon($j2x+I*$j2y) == $j2dragon_x+I*$j2dragon_y\n";
  my $j2dragon_s = $j2dragon_x + $j2dragon_y;
  my $j2dragon_d = $j2dragon_x - $j2dragon_y;

  my %squares;
  my $segments = MyLineSegments->new;
  my $squares = MyLineSegments->new;
  foreach my $n ($n_lo .. $n_hi-1) {
    foreach my $copy (0, 1) {
      my ($x1,$y1) = $path->n_to_xy($n);
      my ($x2,$y2) = $path->n_to_xy($n+1);
      if ($copy) {
        $x1 = $x_hi - $x1;
        $y1 = $y_hi - $y1;
        $x2 = $x_hi - $x2;
        $y2 = $y_hi - $y2;
      }
      $segments->add($x1,$y1, $x2,$y2);

      my $dx = $x2 - $x1;
      my $dy = $y2 - $y1;
      ($dx,$dy) = xy_rotate_plus90($dx,$dy);  # to left side
      $x2 += $dx;
      $y2 += $dy;  # opposite corner
      ($x1,$y1, $x2,$y2) = rect_corners_sort ($x1,$y1, $x2,$y2);
      # $x1,$y1 lower left corner

      my $want_square;
      if ($direction eq 'straight') {
        my $s = $x1 + $y1;
        my $d = $x1 - $y1;
        $want_square = ($s == $j2dragon_s || $d == $j2dragon_d
                        || $s == $jdragon_s || $d == $jdragon_d);
      } else { # $direction eq 'diagonal'
        $want_square = ($x1 == $j2dragon_x || $y1 == $j2dragon_y
                        || $x1 == $jdragon_x || $y1 == $jdragon_y);
      }
      if ($want_square) {
        my ($tx,$ty) = $transform->($x1,$y1);
        $squares{"$tx,$ty"} = 1;
      }
    }
  }

  {
    $Printwrap = 2;
    printwrap("\\fill[my square colour] foreach \\p in {");
    $Printwrap += 2;
    foreach my $xy (sort keys %squares) {
      printwrap("($xy),");
    }
    printwrap("}\n");
    printwrap("{ \\p -- +(1,-1) -- +(0,-2) -- +(-1,-1) };\n");
  }

  $Printwrap = 2;
  printwrap("\\draw ");
  $Printwrap += 2;
  $segments->printwrap (transform => $transform);
  printwrap(";\n");

  {
    my ($x,$y) = $transform->($jdragon_x+.5, $jdragon_y+.5);
    print "  \\node (PlusOffsetJ) at ($x,$y) [my dot];\n";
  }
  {
    my ($x,$y) = $transform->($j2dragon_x+.5, $j2dragon_y+.5);
    print "  \\node (PlusOffsetJ2) at ($x,$y) [my dot];\n";
  }

  print "  \\end{scope}\n";
}

print "  % end generated\n";
exit 0;
