#!/usr/bin/perl -w

# Copyright 2014, 2015, 2016, 2017 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.010;
use strict;
use warnings;
use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()

use FindBin;
use lib $FindBin::Bin;
use MyDragon 'BlobN','BlobNend','xy_rotate_plus90','xy_rotate_minus90';

# uncomment this to run the ### lines
# use Smart::Comments;


my $max_level = 10;
if (@ARGV) { $max_level = shift @ARGV; }
$|=1;

print "% generated by tools/$FindBin::Script $max_level\n";
my $path = Math::PlanePath::DragonCurve->new;

sub path_n_is_right_boundary {
  my ($path, $n, $n_limit) = @_;
  my ($x1,$y1) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  ($dx,$dy) = xy_rotate_minus90($dx,$dy);  # to right side
  my $x1s = $x1 + $dx;
  my $y1s = $y1 + $dy;
  my $x2s = $x2 + $dx;
  my $y2s = $y2 + $dy;
  return (! path_xyxy_is_traversed_either($path, $x1,$y1, $x1s,$y1s, $n_limit)
          || ! path_xyxy_is_traversed_either($path, $x2,$y2, $x2s,$y2s, $n_limit)
          || ! path_xyxy_is_traversed_either($path, $x1s,$y1s, $x2s,$y2s, $n_limit));
}
sub path_n_is_left_boundary {
  my ($path, $n, $n_limit) = @_;
  my ($x1,$y1) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  ($dx,$dy) = xy_rotate_plus90($dx,$dy);  # to side
  my $x1s = $x1 + $dx;
  my $y1s = $y1 + $dy;
  my $x2s = $x2 + $dx;
  my $y2s = $y2 + $dy;
  ### consider: "$x1,$y1 to $x2,$y2   and $x1s,$y1s to $x2s,$y2s"
  return (! path_xyxy_is_traversed_either($path, $x1,$y1, $x1s,$y1s, $n_limit)
          || ! path_xyxy_is_traversed_either($path, $x2,$y2, $x2s,$y2s, $n_limit)
          || ! path_xyxy_is_traversed_either($path, $x1s,$y1s, $x2s,$y2s, $n_limit));
}
sub path_xyxy_is_traversed_either {
  my ($path, $x1,$y1, $x2,$y2, $n_limit) = @_;
  my $n = $path->xyxy_to_n_either($x1,$y1, $x2,$y2);
  return defined $n && $n < $n_limit;
}

my @count;
foreach my $k (0 .. $max_level) {
# foreach my $k (7) {
  ### $k
  my ($n_lo, $n_hi) = $path->level_to_n_range($k);

  my @n_is_right_boundary;
  my $n_is_right_boundary = sub {
    my ($n) = @_;
    return $n_is_right_boundary[$n] //= do {
      my ($x,$y) = $path->n_to_xy($n);
      my ($dx,$dy) = $path->n_to_dxdy($n);
      path_n_is_right_boundary($path, $n, $n_hi) ? 1 : 0
    };
  };

  my @n_is_left_boundary;
  my $n_is_left_boundary = sub {
    my ($n) = @_;
    return $n_is_left_boundary[$n] //= do {
      ### test left: "n=$n  within n_hi=$n_hi"
      my $ret = (path_n_is_left_boundary($path, $n, $n_hi) ? 1 : 0);
      ### test left: "n=$n is $ret"
      $ret
    };
  };

  foreach my $part ('whole','blob') {
    my ($n_start,$n_end);
    if ($part eq 'blob') {
      $n_start = ($k<4 ? 0 : BlobN($k));
      $n_end   = ($k<4 ? 0 : BlobNend($k));
    } else {
      $n_start = $n_lo;
      $n_end   = $n_hi;
    }

    my $n_is_hanging_len = sub {
      my ($n, $length, $boundary_func) = @_;
      if ($n + $length > $n_end) { return 0; }
      foreach my $i (0 .. $length-1) {
        $boundary_func->($n+$i) || return 0;
      }
      my ($x,$y) = $path->n_to_xy($n);
      my @n_list = $path->xy_to_n_list($x,$y);
      ### @n_list
      return (scalar(@n_list) == 2
              && $n_list[0] == $n
              && $n_list[1] == $n+$length);
    };

    foreach my $side ('left','right') {
      my $boundary_func = ($side eq 'left'
                           ? $n_is_left_boundary
                           : $n_is_right_boundary);

      foreach my $length (4, 8, 16) {
        foreach my $n ($n_start .. $n_end-1) {
          if ($n_is_hanging_len->($n, $length, $boundary_func)) {
            ### hanging: "$side n=$n"
            $count[$k]->{$part}->{$side}->[$length]++;
          }
        }
      }

    HQA: for (my $n = $n_start; $n < $n_end; ) {
        foreach my $length (16, 8, 4) {
          if ($n_is_hanging_len->($n, $length, $boundary_func)) {
            ### HQA step: "$part $side n=$n  add ".($length/4)
            $count[$k]->{$part}->{$side}->[0] += $length/4;
            $n += $length;
            next HQA;
          }
        }
        $n++;
      }
    }
  }
}

foreach my $part ('whole','blob') {
  foreach my $side ('both','left','right') {
    foreach my $length (4, 8, 16, 0) {

      print('% GP-DEFINE  ',
            ($part eq 'whole' ? '' : 'Blob'),
            'HQ',
            ($length == 4 ? '1'
             : $length == 8 ? '2'
             : $length == 16 ? '4'
             : 'A'),
            ($side eq 'left' ? 'L'
             : $side eq 'right' ? 'R'
             : ''),
            '_generated_samples = [');
      my $join = '';
      foreach my $k (0 .. $#count) {
        my $value;
        if ($side eq 'both') {
          $value = (($count[$k]->{$part}->{'left'}->[$length] || 0)
                    + ($count[$k]->{$part}->{'right'}->[$length] || 0));
        } else {
          $value = $count[$k]->{$part}->{$side}->[$length] || 0;
        }

        print "$join$value";
        $join = ',';
      }
      print "];\n";
    }
  }
}

print "% end generated\n";
exit 0;
