#!/usr/bin/perl -w

# Copyright 2014, 2015, 2016, 2017, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.004;
use strict;
use Math::PlanePath::DragonCurve 118; # v.118 for xyxy_to_n_either()

use FindBin;
use lib $FindBin::Bin;
use MyDragon 'JA','JN','RQ','chamfer',
  'xy_rotate_plus90','rect_corners_sort','@dir4_to_dx','@dir4_to_dy';
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 10;
my @lengths = (2, 4, 7);
if (@ARGV) { $level = shift @ARGV; }

print "  % generated by tools/$FindBin::Script\n";
print "  \\newcommand{\\MyLevel}{$level}\n";

sub transform {
  my ($x,$y) = @_;
  # no transformation seems best
  return ($x,$y);
}

sub path_n_is_right_boundary {
  my ($path, $n, $n_limit) = @_;
  my ($x1,$y1) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  ($dx,$dy) = xy_rotate_minus90($dx,$dy);  # to side
  my $x1s = $x1 + $dx;
  my $y1s = $y1 + $dy;
  my $x2s = $x2 + $dx;
  my $y2s = $y2 + $dy;
  return (! path_xyxy_is_traversed_either($path, $x1,$y1, $x1s,$y1s, $n_limit)
          || ! path_xyxy_is_traversed_either($path, $x2,$y2, $x2s,$y2s, $n_limit)
          || ! path_xyxy_is_traversed_either($path, $x1s,$y1s, $x2s,$y2s, $n_limit));
}
sub path_n_is_left_boundary {
  my ($path, $n, $n_limit) = @_;
  my ($x1,$y1) = $path->n_to_xy($n);
  my ($x2,$y2) = $path->n_to_xy($n+1);
  my ($dx,$dy) = $path->n_to_dxdy($n);
  ($dx,$dy) = xy_rotate_plus90($dx,$dy);  # to side
  my $x1s = $x1 + $dx;
  my $y1s = $y1 + $dy;
  my $x2s = $x2 + $dx;
  my $y2s = $y2 + $dy;
  return (! path_xyxy_is_traversed_either($path, $x1,$y1, $x1s,$y1s, $n_limit)
          || ! path_xyxy_is_traversed_either($path, $x2,$y2, $x2s,$y2s, $n_limit)
          || ! path_xyxy_is_traversed_either($path, $x1s,$y1s, $x2s,$y2s, $n_limit));
}
sub path_xyxy_is_traversed_either {
  my ($path, $x1,$y1, $x2,$y2, $n_limit) = @_;
  my $n = $path->xyxy_to_n_either($x1,$y1, $x2,$y2);
  return defined $n && $n < $n_limit;
}

my $path = Math::PlanePath::DragonCurve->new;
my ($n_lo, $n_hi) = $path->level_to_n_range($level);

{
  my ($x,$y) = $path->n_to_xy($n_lo);
  ($x,$y) = transform($x,$y);
  print "  \\coordinate (start) at ($x,$y);\n";
}
{
  my ($x,$y) = $path->n_to_xy($n_hi);
  ($x,$y) = transform($x,$y);
  print "  \\coordinate (middle) at ($x,$y);\n";
}
{
  my ($x,$y) = $path->n_to_xy(2*$n_hi);
  ($x,$y) = transform($x,$y);
  print "  \\coordinate (end) at ($x,$y);\n";
}


# {
#   # curve segments
#   require MyLineSegments;
#   my $segments = MyLineSegments->new;
#   foreach my $n (JN($level)-1 .. $n_hi-1) {
#     $segments->add($path->n_to_xy($n),
#                    $path->n_to_xy($n+1));
#   }
#   $Printwrap = 2;
#   printwrap("\\draw ");
#   $Printwrap = 4;
#   $segments->printwrap (transform => \&transform);
#   printwrap(";\n");
# }

# JA(k) = RQ(k-3) + RQ(k-6) + ... RQ(s)
my $squares;
my %seen;
my $k = (($level-3) % 3) - 3;
my $rq = 0;
my @fill_lists;
my @seg_starts;
my @rq_strs;

{
  $Printwrap = 2;
  my $sep = '';
  $Printwrap = 4;
  my @fill_list;

  foreach my $n (JN($level) .. $n_hi - 1) {
    next unless path_n_is_left_boundary($path,$n, $n_hi);
    my ($x1,$y1) = $path->n_to_xy($n);
    my ($dx,$dy) = $path->n_to_dxdy($n);

    my $x2 = $x1 + $dx;
    my $y2 = $y1 + $dy;
    ($dx,$dy) = xy_rotate_plus90($dx,$dy);  # to left side
    $x2 += $dx;
    $y2 += $dy;  # opposite corner

    ($x1,$y1, $x2,$y2) = rect_corners_sort ($x1,$y1, $x2,$y2);
    # ### corners: "$x1,$y1  $x2,$y2"
    next if $seen{"$x1,$y1"}++;

    if ($rq <= 0) {
      if (@fill_list) { push @fill_lists, [@fill_list]; }
      $k += 3;
      # last if $k > $level-2;
      $rq = RQ($k);
      push @rq_strs, "RQ($k)=$rq";
      @fill_list = ();
      push @seg_starts, $n_hi;
    }
    push @fill_list, "$x1,$y1";
    $seg_starts[-1] = $n;
    $rq--;
    $squares++;
  }

  if (@fill_list) {
    ### end fill_list: scalar(@fill_list)
    push @fill_lists, [@fill_list];
  }
  ### end: "rq = $rq"
  # if ($rq) { die "end with rq=$rq"; }
}

my $JA = JA($level);
$squares == $JA
  or warn "squares $squares != JA $JA";

{
  # dashed segment sections
  print "  % segment starts ",join(',',@seg_starts),"  n_hi=$n_hi\n";
  $Printwrap = 2;
  printwrap("\\draw[dashed] ");
  my $join = '';
  foreach my $i (0 .. $#seg_starts) {
    my $seg_start = $seg_starts[$i];
    my $length    = $lengths[$i] || 1;
    $Printwrap = 4;
    foreach my $n (reverse $seg_start-$length+1 .. $seg_start+1) {
      my ($x,$y) = $path->n_to_xy($n);
      printwrap("$join($x,$y)");
      $join = '--';
    }
    $join = ' ';
  }
  printwrap(";\n");
}

foreach my $fill_list (@fill_lists) {
  ### fill_list: join('  ',@$fill_list)
  next unless @$fill_list;

  $Printwrap = 2;
  printwrap("\\fill ");
  $Printwrap = 4;
  my $join = '';

  my @points;
  my @points_dir;
  my @points_gap;
  my @prev_corners;

 P: foreach my $p (@$fill_list) {
    ### $p
    my ($x,$y) = split /,/, $p;
    my @corners = ($x   .','. $y     ,
                   ($x+1).','. $y    ,
                   ($x+1).','.($y+1) ,
                   $x   .','.($y+1)  ,
                  );
    foreach my $corners_pos (0 .. $#corners) {
      foreach my $prev_pos (0 .. $#prev_corners) {
        if ($corners[$corners_pos] eq $prev_corners[$prev_pos]) {
          ### $corners_pos
          ### $prev_pos

          foreach my $points_pos (reverse 0 .. $#points) {
            if ($points[$points_pos] eq $corners[$corners_pos]) {
              ### shuffle points ...
              unshift @points,     (splice @points,     $points_pos);
              unshift @points_dir, (splice @points_dir, $points_pos);
              unshift @points_gap, (splice @points_gap, $points_pos);
              push @points,     $points[0];
              push @points_dir, $points_dir[0];
              push @points_gap, ($prev_pos+1)%4;
              $points_gap[0] == -1 or die;
              $points_gap[0] = $prev_pos;
              last;
            }
          }
          foreach my $i (0 .. $#corners + 1) {
            my $pos = ($corners_pos + $i) % scalar(@corners);
            push @points,     $corners[$pos];
            push @points_dir, $pos;
          }
          push @points_gap, $corners_pos, -1,-1,-1, ($corners_pos+1)%4;
          @prev_corners = @corners;
          next P;
        }
      }
    }

    push @points,     @corners;
    push @points_dir, 0 .. 3;
    push @points_gap, -1,-1,-1,-1;
    @prev_corners = @corners;
  }
  ### points: scalar(@points)."  ".join('  ',@points)
  ### dirs: scalar(@points_dir)."  ".join(', ',@points_dir)
  ### gaps: scalar(@points_gap)."  ".join(', ',@points_gap)

  foreach my $i (0 .. $#points) {
    my $p = $points[$i];
    my $dir = $points_dir[$i];
    my $gap = $points_gap[$i];
    ### $p
    ### $dir
    ### $gap

    my ($x,$y) = split /,/, $p;
    ($x,$y) = transform($x,$y);

    my $dx = $dir4_to_dx[$dir];
    my $dy = $dir4_to_dy[$dir];
    my ($lx,$ly) = xy_rotate_plus90($dx,$dy);  # to left
    $dx += $lx;
    $dy += $ly;
    ($dx,$dy) = transform($dx,$dy);

    $x .= chamfer($dx, "\\MyMargin");
    $y .= chamfer($dy, "\\MyMargin");

    if ($gap >= 0) {
      my $gx = $dir4_to_dx[$gap];
      my $gy = $dir4_to_dy[$gap];
      ($gx,$gy) = transform($gx,$gy);
      $x .= chamfer($gx, "\\MyGap");
      $y .= chamfer($gy, "\\MyGap");
    }
    printwrap("$join($x,$y)");
    $join = '--';
  }
  printwrap("${join}cycle;\n");
}

print "  % fill lists: ",join(', ',map {scalar(@$_)} @fill_lists),"\n";
print "  % RQ list: ",join(', ',@rq_strs),"\n";
print "  % squares: $squares\n";

print "  % end generated\n";
exit 0;
