#!/usr/bin/perl -w

# Copyright 2016, 2017, 2019, 2020 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.

use 5.010;
use strict;
use List::Util 'min','max';
use Math::PlanePath::DragonCurve 117; # v.117 for level_to_n_range()

use FindBin;
use lib "$FindBin::Bin/../tools";
use MyDragon 'transform_to_horizontal';
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $type = 'x';
if (@ARGV) { $type = $ARGV[0]; }

my %type_to_func = ('x' => sub { my ($f,$x,$y) = @_; return $x; },
                    'y' => sub { my ($f,$x,$y) = @_; return $y; },
                    'x-f' => sub { my ($f,$x,$y) = @_; return $x-$f; },
                   );
my $func = $type_to_func{$type}
  // die "Unrecognised type: \"$type\"";

print "  % generated by tools/$FindBin::Script $type\n";

# blocks like
#     n = 0 1 2 3 4 5 6 7 8
#         \-----/ \-------/
# Each block starts on multiple of 2**$block_width_pow2.
# Final n_hi = 2**$level is included in the last block.
#
# Set $num_blocks_pow2 for how many blocks, so how fine resolution in the
# resulting plot.
#
# Set $block_width_pow2 for how many n in each block, so the accuracy of the
# resulting coordinates.
#
# ENHANCE-ME: Could do something with the convex hull of each sub-curve to
# get more accuracy without big $block_width_pow2.  Would need to think
# about what happens to the hull for x-f.

my $num_blocks_pow2 = 10;
my $block_width_pow2 = 8;

my $level = $num_blocks_pow2 + $block_width_pow2;
$level % 2 == 0 or die "prefer level even";
my $scale = 2**($level/2);

my $path = Math::PlanePath::DragonCurve->new;
my ($n_lo,$n_hi) = $path->level_to_n_range($level);

sub n_to_block {
  my ($n) = @_;
  return ($n - ($n==$n_hi)) >> $block_width_pow2;
}

my @t_lo;
my @t_hi;
foreach my $n ($n_lo .. $n_hi) {
  my ($x,$y) = $path->n_to_xy($n);
  ($x,$y) = transform_to_horizontal($x,$y, $level);
  my $block = n_to_block($n);
  my $f = $n/$scale/$scale;
  $x /= $scale;
  $y /= $scale;
  my $t = $func->($f,$x,$y);
  $t_lo[$block] = min($t_lo[$block]//$t, $t);
  $t_hi[$block] = max($t_hi[$block]//$t, $t);
  ### term: "n=$n block=$block  f=$f t=$t  to $t_lo[$block]..$t_hi[$block]"
}
scalar(@t_lo) == 1<<$num_blocks_pow2 or die;
scalar(@t_hi) == 1<<$num_blocks_pow2 or die;

foreach my $t (@t_lo, @t_hi) {
  if ($t) {
    $t = sprintf '%.6f', $t;  # mutate arrays
    $t =~ s/(\.\d)0+/$1/;
    $t =~ s/^0(\.\d)/$1/;
  }
}

# avoid TeX dimension overflow
my $h_scale_pow2 = min(8, $num_blocks_pow2);
my $h_scale_mul = 1/2**($num_blocks_pow2 - $h_scale_pow2);

print "  \\begin{scope}[xscale={1/2^$h_scale_pow2}]\n";
$Printwrap = 2;
printwrap("\\fill[my $type,shift={(line-width-lo)}]");
$Printwrap += 2;

my $join = ' ';
foreach my $block (0 .. $#t_lo) {
  my $f1 = ($block  )*$h_scale_mul;
  my $f2 = ($block+1)*$h_scale_mul;
  printwrap("$join($f1,$t_lo[$block])");
  printwrap("--($f2,$t_lo[$block])");
  $join = '--';
}
printwrap_newline();
printwrap(" [shift={(line-width-hi)}]");

$join = ' --';
foreach my $block (reverse 0 .. $#t_hi) {
  my $f1 = ($block  )*$h_scale_mul;
  my $f2 = ($block+1)*$h_scale_mul;
  printwrap("$join($f2,$t_hi[$block])");
  printwrap("--($f1,$t_hi[$block])");
  $join = '--';
}
printwrap("--cycle;\n");
print "  \\end{scope}\n";

print "  % end generated\n";
exit 0;
